#!/usr/bin/env python
"""
Target videos for the initial batch of hits:

2018-03-15.10-40-00.10-45-00.school.G422.avi
    Person - Texting On Phone: 10
    Vehicle - Turning Left:0
    Person Enters Through Structure:8
    Person Opens Trunk: 0
    People Talking: 18

2018-03-15.10-40-00.10-45-00.bus.G509.avi
    Person - Texting On Phone: 0
    Vehicle - Turning Left:0
    Person Enters Through Structure:3
    Person Opens Trunk: 0
    People Talking: 1

2018-03-15.10-40-01.10-45-01.bus.G327.avi
    Person - Texting On Phone: 21
    Vehicle - Turning Left: 0
    Person Enters Through Structure: 7
    Person Opens Trunk: 0
    People Talking: 7

2018-03-15.10-40-01.10-45-01.bus.G333.avi
    Person - Texting On Phone: 2
    Vehicle - Turning Left: 0
    Person Enters Through Structure: 0
    Person Opens Trunk: 0
    People Talking: 2
"""
import json
import os
import time

import requests

# hit options
sandbox = True
delay = 2  # delay between posting tasks for AWS rate limiting
title = 'Determine activities present in short video clips'
description = (
    'Watch a short video clip and answer which of a given list '
    'of activities occur in the clip.  A large screen is recommended '
    'because videos cannot currently be zoomed.  Also, Safari,'
    'Internet Explorer, and Edge are not supported.  We recommend '
    'Firefox or Chrome on a desktop or laptop computer.'
)
max_assignments = 5
duration = 10 * 60
lifetime = 12 * 60 * 60
reward = 10

# hit definition
videos = [
    '8173918f-8cdc-4a46-ab92-3c9a8c22a9a9',  # 2018-03-15.10-40-00.10-45-00.school.G422.avi
    '17323ad5-dd42-4f31-8884-4aa1d22da06e',  # 2018-03-15.10-40-00.10-45-00.bus.G509.avi
    'e721cd9c-e2cb-40d1-a53d-6d3097a0a1a9',  # 2018-03-15.10-40-01.10-45-01.bus.G327.avi
    '417f3948-88c5-4ebd-adb7-c0c298ab2606'  # 2018-03-15.10-40-01.10-45-01.bus.G333.avi
]
# The activity names should be consistent across hits.  They are
# automatically normalized into lower case when submitted.
activities = json.dumps({
    'person texting on phone': '',  # Could include descriptions when added to the app
    'vehicle turning left': '',
    'person enters through structure': '',
    'person opens trunk': '',
    'people talking': ''
})

app = '2d755948-af16-40bf-846e-0ce885a43b53'
token = os.environ['STUMPF_TOKEN']


def generate_payload(start, end):
    return {
        'sandbox': sandbox,
        'args': {
            'start': start,
            'end': end,
            'activities': activities
        },
        'title': title,
        'description': description,
        'max_assignments': max_assignments,
        'assignment_duration_in_seconds': duration,
        'lifetime_in_seconds': lifetime,
        'reward': reward
    }


def generate_hit(video, start, end):
    return (
        f'https://stumpf.kitware.com/api/v1/task/{video}/{app}/hit',
        generate_payload(start, end)
    )


def serialize_hits():
    with open('source_hits.jsonl', 'w') as f:
        for video in videos:
            for clip in range(15):
                start = clip * 20
                end = start + 20
                f.write(json.dumps(generate_hit(video, start, end)) + '\n')


def submit_hit(hit):
    resp = requests.post(hit[0], json=hit[1], headers={'X-Stumpf-Token': token})
    if not resp.ok:
        try:
            print(resp.json())
        except Exception:
            print(resp.text)
    assert resp.ok


def submit_hits():
    with open('source_hits.jsonl', 'r') as source, \
            open('submitted_hits.jsonl', 'w') as finished, \
            open('failed_hits.jsonl', 'w') as failed:

        for line in source.readlines():
            hit = json.loads(line.strip())
            try:
                submit_hit(hit)
                finished.write(line)
            except Exception:
                failed.write(line)
            finished.flush()
            failed.flush()
            time.sleep(delay)


if __name__ == '__main__':
    serialize_hits()
    submit_hits()
