from datetime import datetime
import uuid

from marshmallow import fields, post_dump, validate, validates, ValidationError
from psycopg2.extras import NumericRange
from sqlalchemy.dialects import postgresql as pg

from stumpf.models import BaseSchema, db

TRACK_TYPES = ['Person', 'Vehicle', 'Sky', 'Building', 'Shadows', 'Grass', 'Ground', 'Trees',
               'Parking_Lot', 'Street', 'Sidewalk', 'Crosswalk', 'Intersection', 'Stairs',
               'Background', 'Unannotated', 'Door', 'Tree', 'Receptacle', 'Dumpster',
               'Parking_Meter', 'ATM', 'Umbrella', 'Construction_Barrier', 'Bike',
               'Construction_Vehicle', 'Push_Pulled_Object', 'Prop', 'Animal',
               'Articulated_Infrastructure', 'Other']

ACTIVITY_TYPES = ['Sit_Down', 'Stand_Up', 'Carry_in_Hands', 'Carry_on_Back', 'Talk_on_Phone',
                  'Text_On_Phone', 'Read_Document', 'Vehicle_Moving', 'Vehicle_Stopping',
                  'Vehicle_Starting', 'Vehicle_Turning_Left', 'Vehicle_Turning_Right',
                  'Vehicle_UTurn', 'Object_Transfer', 'Pick_Up_Object', 'Set_Down_Object',
                  'Heavy_Carry', 'Hand_Interaction', 'Embrace_Interaction', 'Open_Facility_Door',
                  'Close_Facility_Door', 'Enter_Facility', 'Exit_Facility', 'Open_Vehicle_Door',
                  'Close_Vehicle_Door', 'Enter_Vehicle', 'Exit_Vehicle', 'Open_Trunk',
                  'Close_Trunk', 'Load_Vehicle', 'Unload_Vehicle', 'Vehicle_PicksUp_Person',
                  'Vehicle_DropsOff_Person', 'People_Talking', 'Joining_Queue', 'Riding',
                  'Purchasing', 'Laptop_Interaction', 'Abandon_Package', 'Casing_Facility', 'Theft']


class Track(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    type_ = db.Column(pg.ENUM(*TRACK_TYPES, name='track_type_enum'), nullable=False)

    video = db.relationship('Video')


class TrackSchema(BaseSchema):
    __model__ = Track

    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    type_ = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES))
    detections = \
        fields.List(fields.Nested('DetectionSchema', exclude=('video', 'track', )), dump_only=True)


class Range(fields.List):
    def _deserialize(self, value, attr, data):
        a, b = super()._deserialize(value, attr, data)
        return NumericRange(a, b - 1, bounds='[]')

    def _serialize(self, value, attr, obj):
        return super()._serialize([value.lower, value.upper], attr, obj)


class Activity(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    type_ = db.Column(pg.ENUM(*ACTIVITY_TYPES, name='activity_type_enum'), nullable=False)
    frame_range = db.Column(pg.INT4RANGE, nullable=False)

    video = db.relationship('Video')
    tracks = db.relationship('Track', secondary='activity_track')


class ActivitySchema(BaseSchema):
    __model__ = Activity

    video = fields.Nested('VideoSchema', dump_only=True)
    video_id = fields.UUID(required=True, load_only=True)
    type_ = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    frame_range = Range(fields.Int(), required=True)
    tracks = fields.List(fields.Nested('TrackSchema', exclude=('video', )), dump_only=True)

    @validates('frame_range')
    def validate_frame_range(self, frame_range):
        # todo: validate against video frame count
        if frame_range.lower >= frame_range.upper:
            raise ValidationError('Frame range must be at least 1 frame long')


class ActivityGroupSchema(BaseSchema):
    __model__ = Activity

    type_ = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    frame_range = Range(fields.Int(), required=True)
    tracks = fields.List(fields.Nested('TrackSchema', exclude=('video', )), dump_only=True)

    @validates('frame_range')
    def validate_frame_range(self, frame_range):
        # todo: validate against video frame count
        if frame_range.lower >= frame_range.upper:
            raise ValidationError('Frame range must be at least 1 frame long')

    @post_dump
    def form_activity_group(self, data):
        activity_group = {
            'type': data['type_'],
            'activity': {
                'start_frame': data['frame_range'][0],
                'end_frame': data['frame_range'][1] - 1
            },
            'tracks': []
        }
        for track in data['tracks']:
            activity_group_track = {
                'type': track['type_'],
                'detections': []
            }
            for detection in track['detections']:
                activity_group_track['detections'].append({
                   'bounds': detection['bounds'],
                   'source': detection['source'],
                   'frame': detection['frame']
                })
            activity_group['tracks'].append(activity_group_track)
        return activity_group


class ActivityTrack(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    activity_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('activity.id'), nullable=False)
    track_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('track.id'), nullable=False)

    activity = db.relationship('Activity')
    track = db.relationship('Track')


class Detection(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    track_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('track.id'), nullable=False)
    source = db.Column(pg.ENUM('truth', 'ground-truth', 'linear-interpolation',
                               name='detection_source_enum'), nullable=False)
    occlusion = db.Column(pg.ENUM('mostly', 'partially', 'none', name='detection_occlusion_enum'),
                          nullable=False)
    frame = db.Column(db.Integer)
    bounds = db.Column(pg.ARRAY(db.Integer))

    video = db.relationship('Video')
    track = db.relationship('Track', backref=db.backref('detections', lazy=True))


class DetectionSchema(BaseSchema):
    __model__ = Detection

    video = fields.Nested('VideoSchema', dump_only=True)
    video_id = fields.UUID(required=True, load_only=True)
    track = fields.Nested('TrackSchema', dump_only=True)
    track_id = fields.UUID(required=True, load_only=True)
    source = fields.Str(required=True, validate=validate.OneOf(['truth', 'ground-truth',
                                                                'linear-interpolation']))
    occlusion = fields.Str(required=True, validate=validate.OneOf(['mostly', 'partially', 'none']))
    frame = fields.Int(required=True)
    bounds = fields.List(fields.Int(), required=True, validate=validate.Length(min=4, max=4))
