from datetime import datetime, timedelta
import secrets
from typing import Optional
import uuid

from flask_dance.consumer.backend.sqla import OAuthConsumerMixin
from flask_sqlalchemy import SQLAlchemy
from marshmallow import fields, post_load, Schema, validate
from sqlalchemy import DDL, event, MetaData
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.schema import UniqueConstraint


# This is to avoid having to manually name all constraints
# See: http://alembic.zzzcomputing.com/en/latest/naming.html
metadata = MetaData(naming_convention={
    'pk': 'pk_%(table_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'ix': 'ix_%(table_name)s_%(column_0_name)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
})
db = SQLAlchemy(metadata=metadata)
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'))
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "tablefunc";'))


class BaseSchema(Schema):
    __model__ = None

    id = fields.UUID(dump_only=True)
    created = fields.DateTime(dump_only=True)

    @post_load
    def make_object(self, data):
        return self.__model__(**data)


class User(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    email = db.Column(db.String, unique=True, nullable=False)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def is_anonymous(self):
        return not self.is_authenticated

    def get_id(self):
        return self.id


class UserSchema(BaseSchema):
    __model__ = User

    email = fields.Email(required=True)


class OAuth(OAuthConsumerMixin, db.Model):
    provider_user_id = db.Column(db.String(256), unique=True)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(User.id))
    user = db.relationship(User)


def now_plus(**kwargs):
    def f():
        return datetime.utcnow() + timedelta(**kwargs)
    return f


class Token(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = db.Column(db.ForeignKey(User.id, name='token_user_id_fkey'), nullable=False)
    user = db.relationship(User)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False, default=now_plus(days=7))
    token = db.Column(db.String, index=True, unique=True, nullable=False, default=secrets.token_hex)

    @classmethod
    def user_from_token(cls, token: str) -> Optional[User]:
        t = cls.query.filter_by(token=token).filter(cls.expires > datetime.utcnow()).first()

        if t:
            return t.user

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(Token.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_token AS (
                SELECT id, user_id, created, expires
                FROM token)
             """))


class TokenSchema(BaseSchema):
    __model__ = Token

    user_id = fields.UUID(load_only=True)
    user = fields.Nested('UserSchema', dump_only=True)
    expires = fields.DateTime()
    token = fields.Str(dump_only=True)


class MturkToken(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('assignment.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False)
    token = db.Column(db.String, unique=True, nullable=False, default=secrets.token_hex)

    assignment = db.relationship('Assignment', backref=db.backref('mturk_tokens', lazy=True))

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(MturkToken.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_mturk_token AS (
                SELECT id, assignment_id, created, expires
                FROM mturk_token)
             """))


class MturkTokenSchema(BaseSchema):
    __model__ = MturkToken

    expires = fields.DateTime()
    token = fields.Str()
    assignment = fields.Nested('AssignmentSchema', dump_only=True)


class Video(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    path = db.Column(db.String, unique=True, nullable=False)
    name = db.Column(db.String, unique=True, nullable=False)
    frame_rate = db.Column(db.Float, nullable=False)
    duration = db.Column(db.Float, nullable=False)
    width = db.Column(db.Integer, nullable=False)
    height = db.Column(db.Integer, nullable=False)
    camera_gtag = db.Column(db.String, nullable=False)
    start_time = db.Column(db.DateTime(timezone=True), nullable=False)
    end_time = db.Column(db.DateTime(timezone=True), nullable=False)
    node = db.Column(db.String, nullable=False)
    channel = db.Column(db.Integer, nullable=False)
    notes = db.Column(db.Text, nullable=True)

    derived_video_id = db.Column(db.ForeignKey('video.id'), nullable=True)
    derived_video = db.relationship('Video', remote_side=[id],
                                    backref=db.backref('derivations'))

    sortable_fields = {
        'created': created,
        'path': path,
        'name': name,
        'duration': duration,
        'camera_gtag': camera_gtag,
        'channel': channel,
        'start_time': start_time,
        'end_time': end_time
    }
    default_sort = 'name'
    default_desc = False


class VideoSchema(BaseSchema):
    __model__ = Video

    path = fields.Url(required=True)
    name = fields.Str(required=True)
    frame_rate = fields.Float(required=True)
    duration = fields.Float(required=True)
    width = fields.Integer(required=True)
    height = fields.Integer(required=True)
    camera_gtag = fields.Str(required=True)
    start_time = fields.DateTime(required=True)
    end_time = fields.DateTime(required=True)
    node = fields.Str(required=True)
    channel = fields.Integer(required=True)
    notes = fields.Str(required=False)

    derived_video_id = fields.UUID(load_only=True)
    derived_video = fields.Nested('VideoSchema', dump_only=True)


class AnnotationApp(db.Model):
    __table_args__ = (UniqueConstraint('repo_url', 'name', 'url'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    repo_url = db.Column(db.String, nullable=False)
    name = db.Column(db.String, nullable=False)
    url = db.Column(db.String, nullable=False)

    sortable_fields = {
        'created': created,
        'name': name
    }
    default_sort = 'created'
    default_desc = True


class AnnotationAppSchema(BaseSchema):
    __model__ = AnnotationApp

    repo_url = fields.Url(required=True)
    url = fields.Url(required=True)
    name = fields.Str(required=True)


class Task(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('tasks', lazy=True))
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False,
                         index=True)
    video = db.relationship('Video', backref=db.backref('tasks', lazy=True))
    annotation_app_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('annotation_app.id'),
                                  nullable=False)
    annotation_app = db.relationship('AnnotationApp', backref=db.backref('tasks', lazy=True))
    sandbox = db.Column(pg.BOOLEAN, nullable=False, default=False)
    hit_id = db.Column(db.String, nullable=True, unique=True)
    hit_group_id = db.Column(db.String, nullable=True)
    type_ = db.Column(pg.ENUM('HIT', 'KWIT', name='hit_kwit_enum'), nullable=False)
    title = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    assignment_duration_in_seconds = db.Column(db.Integer, nullable=False)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               name='task_status_enum'), nullable=False, default='Assignable')
    review_status = db.Column(pg.ENUM('NotReviewed', 'MarkedForReview', 'ReviewedAppropriate',
                                      'ReviewedInappropriate', name='task_review_status_enum'),
                              nullable=False, default='NotReviewed')
    max_assignments = db.Column(db.Integer, nullable=False)
    reward = db.Column(db.Integer, nullable=True)  # cents
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})
    answer = db.Column(pg.JSONB, nullable=True)

    searchable_fields = {
        'type': type_,
        'task_name': title,
        'assignments': max_assignments
    }

    sortable_fields = {
        'created': created,
        'title': title,
        'type_': type_
    }
    default_sort = 'created'
    default_desc = True


class TaskSchema(BaseSchema):
    __model__ = Task

    annotation_app = fields.Nested('AnnotationAppSchema', dump_only=True)
    title = fields.Str(required=True)
    type_ = fields.Str(required=True)
    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    sandbox = fields.Boolean(required=True)
    hit_id = fields.Str()
    hit_group_id = fields.Str()
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    assignment_duration_in_seconds = fields.Int()
    status = fields.Str()
    review_status = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))


class CreateHitSchema(Schema):
    sandbox = fields.Bool(missing=True)
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    auto_approval_delay_in_seconds = fields.Int(missing=None)
    assignment_duration_in_seconds = fields.Int(required=True,
                                                validate=validate.Range(min=30, max=31536000))
    lifetime_in_seconds = fields.Int(required=True,
                                     validate=validate.Range(min=30, max=31536000))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))
    args = fields.Dict(keys=fields.Str())


class Assignment(db.Model):
    __table_args__ = (UniqueConstraint('task_id', 'assignment_id', 'worker_id'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    assignment_id = db.Column(db.String, nullable=True)
    worker_id = db.Column(db.String, nullable=True)
    app_version = db.Column(db.String, nullable=True, default=None)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)
    answer = db.Column(pg.JSONB, nullable=True)

    task = db.relationship('Task', backref=db.backref('assignments', lazy=True))

    sortable_fields = {
        'created': created
    }
    default_sort = 'created'
    default_desc = True

    @property
    def status_changes(self):
        return AssignmentStatusChange.query.filter_by(assignment_id=self.id) \
                                           .order_by('occurred').all()


event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_times AS (
                SELECT
                assignment_id, accepted, submitted, returned, abandoned, approved,
                EXTRACT(epoch from (submitted - accepted)) as tt_submitted,
                EXTRACT(epoch from (approved - accepted))  as tt_approved,
                EXTRACT(epoch from (returned - accepted))  as tt_returned,
                EXTRACT(epoch from (abandoned - accepted))  as tt_abandoned
                FROM
                CROSSTAB('
                    SELECT cross_product.assignment_id, cross_product.status, s.occurred
                    FROM (
                    SELECT DISTINCT assignment_id, statuses.status
                    FROM assignment_status_change
                    CROSS JOIN (
                        SELECT DISTINCT status FROM assignment_status_change
                        ) AS statuses
                    ) AS cross_product
                    LEFT OUTER JOIN assignment_status_change AS s
                    ON cross_product.assignment_id = s.assignment_id AND
                       cross_product.status = s.status
                    ORDER BY 1, 2
                ') AS (assignment_id UUID,
                    abandoned     TIMESTAMP,
                    returned      TIMESTAMP,
                    accepted      TIMESTAMP,
                    submitted     TIMESTAMP,
                approved TIMESTAMP));
"""))

event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_detail AS (
                SELECT vast.assignment_id,
                    vast.accepted,
                    vast.submitted,
                    vast.returned,
                    vast.abandoned,
                    vast.approved,
                    vast.tt_submitted,
                    vast.tt_approved,
                    vast.tt_returned,
                    vast.tt_abandoned,
                    a.created AS assignment_created,
                    a.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds,
                    v.id AS video_id,
                    v.path,
                    v.frame_rate,
                    v.duration
                FROM view_assignment_status_times vast,
                    assignment a,
                    annotation_app aa,
                    task t,
                    video v
                WHERE a.task_id = t.id AND
                        v.id = t.video_id AND
                        aa.id = t.annotation_app_id AND
                        vast.assignment_id = a.id);
"""))


class AssignmentStatusChange(db.Model):
    __table_args__ = (UniqueConstraint('assignment_id', 'status'),)

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    occurred = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Assignment.id), nullable=False)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)

    assignment = db.relationship('Assignment')


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION update_assignment_status()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                UPDATE assignment SET status = (
                    SELECT status FROM assignment_status_change
                    WHERE assignment_id = NEW.assignment_id
                    ORDER BY occurred DESC LIMIT 1)
                WHERE id = NEW.assignment_id;
             RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER update_assignment_status
             AFTER INSERT OR UPDATE ON assignment_status_change
             FOR EACH ROW EXECUTE PROCEDURE update_assignment_status();"""))


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_change_detail AS (
                SELECT a.id,
                    a.created AS assignment_created,
                    asc_.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds,
                    v.id AS video_id,
                    v.path,
                    v.frame_rate,
                    v.duration
                FROM assignment_status_change asc_,
                    assignment a,
                    annotation_app aa,
                    task t,
                    video v
                WHERE a.task_id = t.id AND
                        v.id = t.video_id AND
                        aa.id = t.annotation_app_id AND
                        asc_.assignment_id = a.id);
"""))


class AssignmentStatusChangeSchema(BaseSchema):
    __model__ = AssignmentStatusChange

    created = fields.DateTime(dump_only=True)
    occurred = fields.DateTime()
    assignment_id = fields.Str()
    status = fields.Str()


class AssignmentSchema(BaseSchema):
    __model__ = Assignment

    task = fields.Nested('TaskSchema', dump_only=True)
    assignment_id = fields.Str()
    worker_id = fields.Str()
    status = fields.Str()
    app_version = fields.Str(required=True)
    status_changes = fields.List(fields.Nested('AssignmentStatusChangeSchema'), dump_only=True)


class Log(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    category = db.Column(db.String, nullable=False, index=True)
    message = db.Column(pg.JSONB, nullable=False, default=lambda: {})


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION log_notify()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                     IF NOT NEW.message ? 'channel' THEN
                        RAISE EXCEPTION 'NOTIFY messages must contain a "channel" attribute';
                     END IF;
                     IF NOT NEW.message ? 'payload' THEN
                        RAISE EXCEPTION 'NOTIFY messages must contain a "payload" attribute';
                     END IF;

                     PERFORM pg_notify(NEW.message ->> 'channel', NEW.message ->> 'payload');
                     RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER log_notify
             BEFORE INSERT ON log
             FOR EACH ROW WHEN (NEW.category = 'NOTIFY')
             EXECUTE PROCEDURE log_notify();"""))


class LogSchema(BaseSchema):
    __model__ = Log

    category = fields.Str(required=True)
    message = fields.Dict(required=True)


class Feedback(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    ip = db.Column(db.String, nullable=False)
    user_agent = db.Column(db.String, nullable=False)
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})


class FeedbackSchema(BaseSchema):
    __model__ = Feedback

    ip = fields.Str(required=True)
    user_agent = fields.Str(required=True)
    args = fields.Dict(required=True)
